#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>
#include <memory.h>

#include <windows.h>

#include <crtdbg.h>

#include <mqoobject.h>

#include <DispObj.h>
#include <polymesh3.h>
#include <GetMaterial.h>

#include <mqoface.h>
#include <mqomaterial.h>

//#include <extlineio.h>
//#include <extline.h>

#include <algorithm>
#include <iostream>
#include <iterator>

static int s_alloccnt = 0;

typedef struct tag_latheelem
{
	float height;
	float dist;
} LATHEELEM;


CMQOObject::CMQOObject()
{
	InitParams();
	s_alloccnt++;
	m_objectno = s_alloccnt;
}

CMQOObject::~CMQOObject()
{
	if( m_pointbuf ){
		free( m_pointbuf );
		m_pointbuf = 0;
	}

	if( m_facebuf ){
		delete [] m_facebuf;
		m_facebuf = 0;
	}

	if( m_colorbuf ){
		free( m_colorbuf );
		m_colorbuf = 0;
	}

	if( m_pointbuf2 ){
		free( m_pointbuf2 );
		m_pointbuf2 = 0;
	}

	if( m_facebuf2 ){
		delete [] m_facebuf2;
		m_facebuf2 = 0;
	}

	if( m_colorbuf2 ){
		free( m_colorbuf2 );
		m_colorbuf2 = 0;
	}

	if( m_connectface ){
		delete [] m_connectface;
		m_connectface = 0;
	}

	if( m_pm3 ){
		delete m_pm3;
		m_pm3 = 0;
	}

	if( m_dispobj ){
		delete m_dispobj;
		m_dispobj = 0;
	}
}

void CMQOObject::InitParams()
{
	m_objectno = -1;
	ZeroMemory( m_name, 256 );

	m_patch = 0;
	m_segment = 1;

	m_visible = 15;
	m_locking = 0;

	m_shading	= 1;
	m_facet = 59.5f;
	m_color.x = 1.0f;
	m_color.y = 1.0f;
	m_color.z = 1.0f;
	m_color.w = 1.0f;
	m_color_type = 0;
	m_mirror = 0;
	m_mirror_axis = 1;
	m_issetmirror_dis = 0;
	m_mirror_dis = 0.0f;
	m_lathe = 0;
	m_lathe_axis = 0;
	m_lathe_seg = 3;
	m_vertex = 0;
	//BVertex;
	m_face = 0;

	m_pointbuf = 0;
	m_facebuf = 0;

	m_vertex2 = 0;
	m_face2 = 0;
	m_pointbuf2 = 0;
	m_facebuf2 = 0;

	m_hascolor = 0;
	m_colorbuf = 0;
	m_colorbuf2 = 0;

	m_connectnum = 0;
	m_connectface = 0;

	m_pm3 = 0;
	D3DXMatrixIdentity( &m_multmat );

	m_dispobj = 0;

//	next = 0;
}


int CMQOObject::SetParams( char* srcchar, int srcleng )
{
	char pat[15][20] = 
	{
		"Object",
		"patch",
		"segment",
		"shading",
		"facet",
		"color_type",
		"color",
		"mirror_axis",
		"mirror_dis",
		"mirror",
		"lathe_axis",
		"lathe_seg",
		"lathe",

		"visible",
		"locking"
	};

	int patno, patleng;
	int pos = 0;
	int cmp;

	int stepnum;
	for( patno = 0; patno < 15; patno++ ){
		while( (pos < srcleng) && ( (*(srcchar + pos) == ' ') || (*(srcchar + pos) == '\t') ) ){
			pos++;
		}
		
		patleng = (int)strlen( pat[patno] );
		cmp = strncmp( pat[patno], srcchar + pos, patleng );
		if( cmp == 0 ){

			pos += patleng;//!!!

			switch( patno ){
			case 0:
				CallF( GetName( m_name, 256, srcchar, pos, srcleng ), return 1 );
				break;
			case 1:
				CallF( GetInt( &m_patch, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 2:
				CallF( GetInt( &m_segment, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 3:
				CallF( GetInt( &m_shading, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 4:
				CallF( GetFloat( &m_facet, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 5:
				CallF( GetInt( &m_color_type, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 6:
				CallF( GetFloat( &m_color.x, srcchar, pos, srcleng, &stepnum ), return 1 );
				pos += stepnum;
				if( pos >= srcleng ){
					DbgOut( L"MQOObject : SetParams : GetFloat : pos error !!!" );
					_ASSERT( 0 );
					return 1;
				}


				CallF( GetFloat( &m_color.y, srcchar, pos, srcleng, &stepnum ), return 1 );
				pos += stepnum;
				if( pos >= srcleng ){
					DbgOut( L"MQOObject : SetParams : GetFloat : pos error !!!" );
					_ASSERT( 0 );
					return 1;
				}

				CallF( GetFloat( &m_color.z, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 7:
				CallF( GetInt( &m_mirror_axis, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 8:
				CallF( GetFloat( &m_mirror_dis, srcchar, pos, srcleng, &stepnum ), return 1 );
				m_issetmirror_dis = 1;//!!!!
				break;
			case 9:
				CallF( GetInt( &m_mirror, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 10:
				CallF( GetInt( &m_lathe_axis, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 11:
				CallF( GetInt( &m_lathe_seg, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 12:
				CallF( GetInt( &m_lathe, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 13:
				CallF( GetInt( &m_visible, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			case 14:
				CallF( GetInt( &m_locking, srcchar, pos, srcleng, &stepnum ), return 1 );
				break;
			default:
				_ASSERT( 0 );
				break;
			}

			break;
		}
	}
	


	return 0;

}

int CMQOObject::GetInt( int* dstint, char* srcchar, int pos, int srcleng, int* stepnum )
{
	int startpos, endpos;
	startpos = pos;

	while( (startpos < srcleng) &&  
		( ( isdigit( *(srcchar + startpos) ) == 0 ) && (*(srcchar + startpos) != '-') ) 
	
	){
		startpos++;
	}

	endpos = startpos;
	while( (endpos < srcleng) && 
		( (isdigit( *(srcchar + endpos) ) != 0) || ( *(srcchar + endpos) == '-' ) )
	){
		endpos++;
	}

	char tempchar[256];
	if( (endpos - startpos < 256) && (endpos - startpos > 0) ){
		strncpy_s( tempchar, 256, srcchar + startpos, endpos - startpos );
		tempchar[endpos - startpos] = 0;

		*dstint = atoi( tempchar );

		*stepnum = endpos - pos;
	}else{
		_ASSERT( 0 );
		*stepnum = endpos - pos;
		return 1;
	}


	return 0;
}
int CMQOObject::GetFloat( float* dstfloat, char* srcchar, int pos, int srcleng, int* stepnum )
{
	int startpos, endpos;
	startpos = pos;

	while( (startpos < srcleng) &&  
		( ( isdigit( *(srcchar + startpos) ) == 0 ) && (*(srcchar + startpos) != '-') && (*(srcchar + startpos) != '.') ) 
	
	){
		startpos++;
	}

	endpos = startpos;
	while( (endpos < srcleng) && 
		( (isdigit( *(srcchar + endpos) ) != 0) || ( *(srcchar + endpos) == '-' ) || (*(srcchar + endpos) == '.') )
	){
		endpos++;
	}

	char tempchar[256];
	if( (endpos - startpos < 256) && (endpos - startpos > 0) ){
		strncpy_s( tempchar, 256, srcchar + startpos, endpos - startpos );
		tempchar[endpos - startpos] = 0;

		*dstfloat = (float)atof( tempchar );

		*stepnum = endpos - pos;
	}else{
		_ASSERT( 0 );
		*stepnum = endpos - pos;
		return 1;
	}


	return 0;
}

int CMQOObject::GetName( char* dstchar, int dstleng, char* srcchar, int pos, int srcleng )
{
	int startpos, endpos;
	startpos = pos;

	while( (startpos < srcleng) && 
		( ( *(srcchar + startpos) == ' ' ) || ( *(srcchar + startpos) == '\t' ) || ( *(srcchar + startpos) == '\"' ) ) 
	){
		startpos++;
	}

	endpos = startpos;
	while( (endpos < srcleng) && 
		( ( *(srcchar + endpos) != ' ' ) && ( *(srcchar + endpos) != '\t' ) && (*(srcchar + endpos) != '\r') && (*(srcchar + endpos) != '\r\n') && (*(srcchar + endpos) != '\"') )
	){
		endpos++;
	}

	if( (endpos - startpos < dstleng) && (endpos - startpos > 0) ){
		strncpy_s( dstchar, dstleng, srcchar + startpos, endpos - startpos );
		*(dstchar + endpos - startpos) = 0;

	}else{
		_ASSERT( 0 );
	}


	return 0;
}

int CMQOObject::SetVertex( int* vertnum, char* srcchar, int srcleng )
{
	//vertex, ܂́A@BVertex܂ޕ󂯎B
	if( m_pointbuf ){
		_ASSERT( 0 );
		return 1;
	}

	int ret;

	char headerchar1[] = "vertex";
	int headerleng1 = (int)strlen( headerchar1 );
	char* find1;
	find1 = strstr( srcchar, headerchar1 );


	char headerchar2[] = "Vector";
	int headerleng2 = (int)strlen( headerchar2 );
	char* find2;
	find2 = strstr( srcchar, headerchar2 );

	int pos;
	if( find1 != NULL ){
		pos = (int)(find1 + headerleng1 - srcchar);
	}else if( find2 != NULL ){
		pos = (int)(find2 + headerleng2 - srcchar);
	}else{

	}
	int stepnum;
	ret = GetInt( &m_vertex, srcchar, pos, srcleng, &stepnum );
	if( ret ){
		_ASSERT( 0 );
		return 1;
	}
	*vertnum = m_vertex;
	DbgOut( L"MQOObject : SetVertex : vertex %d\r\n", m_vertex );


	m_pointbuf = (D3DXVECTOR3*)malloc( sizeof( D3DXVECTOR3 ) * m_vertex );
	if( !m_pointbuf ){
		_ASSERT( 0 );
		return 1;
	}

	ZeroMemory( m_pointbuf, sizeof( D3DXVECTOR3 ) * m_vertex );


	return 0;
}
int CMQOObject::SetPointBuf( unsigned char* srcptr, int srcnum )
{
	if( !m_pointbuf || (srcnum > m_vertex) ){
		_ASSERT( 0 );
		return 1;
	}

	MoveMemory( m_pointbuf, srcptr, srcnum * sizeof( VEC3F ) );

	return 0;

}


int CMQOObject::SetPointBuf( int vertno, char* srcchar, int srcleng )
{
	if( !m_pointbuf || (vertno >= m_vertex) ){
		_ASSERT( 0 );
		return 1;
	}


	D3DXVECTOR3* dstvec;
	dstvec = m_pointbuf + vertno;
	int pos, stepnum;
	pos = 0;

	CallF( GetFloat( &dstvec->x, srcchar, pos, srcleng, &stepnum ), return 1 );
	pos += stepnum;
	if( pos >= srcleng ){
		DbgOut( L"MQOObject : SePointBuf : GetFloat : pos error !!!" );
		_ASSERT( 0 );
		return 1;
	}


	CallF( GetFloat( &dstvec->y, srcchar, pos, srcleng, &stepnum ), return 1 );
	pos += stepnum;
	if( pos >= srcleng ){
		DbgOut( L"MQOObject : SePointBuf : GetFloat : pos error !!!" );
		_ASSERT( 0 );
		return 1;
	}


	CallF( GetFloat( &dstvec->z, srcchar, pos, srcleng, &stepnum ), return 1 );

	return 0;
}

int CMQOObject::SetFace( int* facenum, char* srcchar, int srcleng )
{
	if( m_facebuf ){
		_ASSERT( 0 );
		return 1;
	}

	int ret;
	char headerchar[] = "face";
	int headerleng = (int)strlen( headerchar );

	char* find;
	find = strstr( srcchar, headerchar );
	if( find == NULL ){
		_ASSERT( 0 );
		return 1;
	}

	int pos;
	pos = (int)(find + headerleng - srcchar);
	int stepnum;
	ret = GetInt( &m_face, srcchar, pos, srcleng, &stepnum );
	if( ret ){
		_ASSERT( 0 );
		return 1;
	}
	*facenum = m_face;
	DbgOut( L"MQOObject : SetFace : face %d\r\n", m_face );

	
	m_facebuf = new CMQOFace[ m_face ];
	if( !m_facebuf ){
		_ASSERT( 0 );
		return 1;
	}


	return 0;
}

int CMQOObject::SetFaceBuf( int faceno, char* srcchar, int srcleng, int materialoffset )
{
	if( !m_facebuf || (faceno >= m_face) ){
		_ASSERT( 0 );
		return 1;
	}


	CMQOFace* dstface;
	dstface = m_facebuf + faceno;

	_ASSERT( dstface );

	CallF( dstface->SetParams( srcchar, srcleng ), return 1 );
	dstface->m_materialno += materialoffset;


	return 0;
}

int CMQOObject::Dump()
{
	WCHAR wname[256] = {0};
	MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, m_name, 256, wname, 256 );


	DbgOut( L"Ojbect : %s\r\n", wname );
	DbgOut( L"\tpatch %d\r\n", m_patch );
	DbgOut( L"\tsegment %d\r\n", m_segment );

	DbgOut( L"\tvisible %d\r\n", m_visible );
	DbgOut( L"\tlocking %d\r\n", m_locking );

	DbgOut( L"\tshading %d\r\n", m_shading );
	DbgOut( L"\tfacet %f\r\n", m_facet );
	DbgOut( L"\tcolor %f %f %f\r\n", m_color.x, m_color.y, m_color.z );
	DbgOut( L"\tcolor_type %d\r\n", m_color_type );
	DbgOut( L"\tmirror %d\r\n", m_mirror );
	DbgOut( L"\tmirror_axis %d\r\n", m_mirror_axis );
	DbgOut( L"\tmirror_dis %f\r\n", m_mirror_dis );
	DbgOut( L"\tlathe %d\r\n", m_lathe );
	DbgOut( L"\tlathe_axis %d\r\n", m_lathe_axis );
	DbgOut( L"\tlathe_seg %d\r\n", m_lathe_seg );

	DbgOut( L"\tvertex %d\r\n", m_vertex );
	int vertno;
	D3DXVECTOR3* curvec = m_pointbuf;
	for( vertno = 0; vertno < m_vertex; vertno++ ){
		DbgOut( L"\t\t%f %f %f\r\n", curvec->x, curvec->y, curvec->z );
		curvec++;
	}

	if( m_hascolor && m_colorbuf ){
		DbgOut( L"\tcolor\r\n" );
		for( vertno = 0; vertno < m_vertex; vertno++ ){
			DbgOut( L"\t\t%d : %f %f %f\r\n", vertno, m_colorbuf[vertno].x, m_colorbuf[vertno].y, m_colorbuf[vertno].z );
		}
	}

	int ret;
	DbgOut( L"\tface %d\r\n", m_face );
	int faceno;
	CMQOFace* curface = m_facebuf;
	for( faceno = 0; faceno < m_face; faceno++ ){
		ret = curface->Dump();
		if( ret ){
			_ASSERT( 0 );
			return 1;
		}
		curface++;
	}

	return 0;
}

int CMQOObject::CreateColor()
{
	if( m_vertex <= 0 ){
		DbgOut( L"MQOObject : CreateColor : vertex error !!!\r\n" );
		_ASSERT( 0 );
		return 1;
	}

	m_colorbuf = (D3DXVECTOR4*)malloc( sizeof( D3DXVECTOR4 ) * m_vertex );
	if( !m_colorbuf ){
		DbgOut( L"MQOObject : CreateColor : colorbuf alloc error !!!\r\n" );
		_ASSERT( 0 );
		return 1;
	}

	ZeroMemory( m_colorbuf, sizeof( D3DXVECTOR4 ) * m_vertex );

	m_hascolor = 1;

	return 0;
}
int CMQOObject::SetColor( char* srcchar, int srcleng )
{
	int pos, stepnum;

	int pointno;
	
	pos = 0;
	CallF( GetInt( &pointno, srcchar, pos, srcleng, &stepnum ), return 1 );

	if( (pointno < 0) || (pointno >= m_vertex) ){
		DbgOut( L"MQOObject : SetColor : pointno error !!!\r\n" );
		_ASSERT( 0 );
		return 1;
	}


	pos += stepnum;
	DWORD col;
	CallF( GetInt( (int*)&col, srcchar, pos, srcleng, &stepnum ), return 1 );

	DWORD dwr, dwg, dwb;
	dwr = col & 0x000000FF;
	dwg = (col & 0x0000FF00) >> 8;
	dwb = (col & 0x00FF0000) >> 16;

	D3DXVECTOR4* dstcol = m_colorbuf + pointno;
	dstcol->x = (float)dwr / 255.0f;
	dstcol->y = (float)dwg / 255.0f;
	dstcol->z = (float)dwb / 255.0f;
	dstcol->w = 1.0f;
	

	return 0;
}
int CMQOObject::MakePolymesh3( LPDIRECT3DDEVICE9 pdev, map<int,CMQOMaterial*>& srcmaterial )
{
	if( !m_pointbuf || !m_facebuf )
		return 0;

	if( m_pm3 ){
		delete m_pm3;
		m_pm3 = 0;
	}
	if( m_dispobj ){
		delete m_dispobj;
		m_dispobj = 0;
	}



	
	if( m_lathe != 0 ){
		CallF( MakeLatheBuf(), return 1 );
	}

	if( m_mirror != 0 ){
		CallF( MakeMirrorBuf(), return 1 );
	}
	

////////


	// ʂƒ_̐擾
	int face_count;
	int vert_count;
	D3DXVECTOR3* pointptr;
	CMQOFace* faceptr;
	D3DXVECTOR4* colorptr;

	if( m_face2 > 0 ){
		face_count = m_face2;
		faceptr = m_facebuf2;
	}else{
		face_count = m_face;
		faceptr = m_facebuf;
	}

	if( m_vertex2 > 0 ){
		vert_count = m_vertex2;
		pointptr = m_pointbuf2;
	}else{
		vert_count = m_vertex;
		pointptr = m_pointbuf;
	}

	if( m_colorbuf2 ){
		colorptr = m_colorbuf2;
	}else{
		colorptr = m_colorbuf;
	}

	m_pm3 = new CPolyMesh3();
	if( !m_pm3 ){
		_ASSERT( 0 );
		return 1;
	}
	CallF( m_pm3->CreatePM3( vert_count, face_count, m_facet, pointptr, faceptr, srcmaterial, m_multmat ), return 1 );

	if( m_pm3->m_createoptflag ){
		m_dispobj = new CDispObj();
		if( !m_dispobj ){
			_ASSERT( 0 );
			return 1;
		}
		CallF( m_dispobj->CreateDispObj( pdev, m_pm3 ), return 1 );
	}

	return 0;
}
int CMQOObject::HasPolygon()
{
	int face2 = 0;
	int face3 = 0;
	int face4 = 0;

	int faceno;

	for( faceno = 0; faceno < m_face; faceno++ ){
		CMQOFace* curface;
		curface = m_facebuf + faceno;
		switch( curface->m_pointnum ){
			case 2:
				face2++;
				break;
			case 3:
				face3++;
				break;
			case 4:
				face4++;
			default:
				break;
		}
	}
	
	if( m_lathe != 0 ){
		return (face2 + face3 + face4);
	}else{
		return (face3 + face4);
	}
}

int CMQOObject::HasLine()
{
	int face2 = 0;

	int faceno;

	for( faceno = 0; faceno < m_face; faceno++ ){
		CMQOFace* curface;
		curface = m_facebuf + faceno;
		switch( curface->m_pointnum ){
			case 2:
				face2++;
				break;
			case 3:
			case 4:
			default:
				break;
		}
	}
	
	return face2;
}

/***
int CMQOObject::IsMikoBone()
{
	int cmp;
	char pattern[20] = "bone:";
	int patleng = (int)strlen( pattern );

	cmp = strncmp( m_name, pattern, patleng );
	if( cmp == 0 )
		return 1;
	else
		return 0;
}
***/


int CMQOObject::MakeLatheBuf()
{
	int linenum = 0;
	int faceno;
	CMQOFace* curface;
	for( faceno = 0; faceno < m_face; faceno++ ){
		curface = m_facebuf + faceno;
		if( curface->m_pointnum == 2 ){
			linenum++;
		}
	}

	if( linenum <= 0 ) //!!!!
		return 0;


	int* lineno2faceno;
	lineno2faceno = (int*)malloc( sizeof( int ) * linenum );
	if( !lineno2faceno ){
		DbgOut( L"MQOObject : MakeLatheBuf : lineno2faceno alloc error !!!\r\n" );
		_ASSERT( 0 );
		return 1;
	}
	ZeroMemory( lineno2faceno, sizeof( int ) * linenum );

	int lineno = 0;
	for( faceno = 0; faceno < m_face; faceno++ ){
		curface = m_facebuf + faceno;
		if( curface->m_pointnum == 2 ){
			_ASSERT( lineno < linenum );
			*(lineno2faceno + lineno) = faceno; 
			lineno++;
		}
	}

	if( m_pointbuf2 ){
		_ASSERT( 0 );
		return 1;
	}
	m_vertex2 = m_lathe_seg * 2 * linenum;
	m_pointbuf2 = (D3DXVECTOR3*)malloc( sizeof( D3DXVECTOR3 ) * m_vertex2 );
	if( !m_pointbuf2 ){
		DbgOut( L"MQOObject : MakeLatheBuf : pointbuf2 alloc error !!!\r\n" );
		_ASSERT( 0 );
		return 1;			
	}
	ZeroMemory( m_pointbuf2, sizeof( D3DXVECTOR3 ) * m_vertex2 );

	if( m_colorbuf ){
		if( m_colorbuf2 ){
			_ASSERT( 0 );
			return 1;
		}
		m_colorbuf2 = (D3DXVECTOR4*)malloc( sizeof( D3DXVECTOR4 ) * m_vertex2 );
		if( !m_colorbuf2 ){
			DbgOut( L"MQOObject : MakeLatheBuf : colorbuf2 alloc error !!!\r\n" );
			_ASSERT( 0 );
			return 1;			
		}
		//ZeroMemory( colorbuf2, sizeof( ARGBF ) * vertex2 );
		MoveMemory( m_colorbuf2, m_colorbuf, sizeof( D3DXVECTOR4 ) * m_vertex );
	}

	if( m_facebuf2 ){
		_ASSERT( 0 );
		return 1;
	}
	m_face2 = m_lathe_seg * linenum;
	m_facebuf2 = new CMQOFace[ m_face2 ];
	if( !m_facebuf2 ){
		DbgOut( L"MQOObject : MakeLatheBuf : facebuf2 alloc error !!!\r\n" );
		_ASSERT( 0 );
		return 1;			
	}
	

	LATHEELEM elem[2];
	float mag, rad;
	int segno;
	rad = 2.0f * PI / m_lathe_seg;

	for( lineno = 0; lineno < linenum; lineno++ ){
		faceno = *(lineno2faceno + lineno);
		int v0, v1;
		v0 = (m_facebuf + faceno)->m_index[0];
		v1 = (m_facebuf + faceno)->m_index[1];
		D3DXVECTOR3* src0 = m_pointbuf + v0;
		D3DXVECTOR3* src1 = m_pointbuf + v1;
		switch( m_lathe_axis ){
		case 0://X
			elem[0].height = src0->x;
			mag = src0->y * src0->y + src0->z * src0->z;
			if( mag > 0.0f ){
				elem[0].dist = (float)sqrt( mag );
			}else{
				elem[0].dist = 0.0f;
			}

			elem[1].height = src1->x;
			mag = src1->y * src1->y + src1->z * src1->z;
			if( mag > 0.0f ){
				elem[1].dist = (float)sqrt( mag );
			}else{
				elem[1].dist = 0.0f;
			}

			for( segno = 0; segno < m_lathe_seg; segno++ ){
				D3DXVECTOR3* dst0 = m_pointbuf2 + lineno * 2 * m_lathe_seg + segno;
				D3DXVECTOR3* dst1 = dst0 + m_lathe_seg;

				dst0->x = elem[0].height;
				dst0->y = elem[0].dist * (float)cos( rad * segno );
				dst0->z = elem[0].dist * (float)sin( rad * segno );

				dst1->x = elem[1].height;
				dst1->y = elem[1].dist * (float)cos( rad * segno );
				dst1->z = elem[1].dist * (float)sin( rad * segno );

				if( m_colorbuf2 ){
					D3DXVECTOR4* dstcol0 = m_colorbuf2 + lineno * 2 * m_lathe_seg + segno;
					D3DXVECTOR4* dstcol1 = dstcol0 + m_lathe_seg;

					*dstcol0 = *(m_colorbuf + v0);
					*dstcol1 = *(m_colorbuf + v1);
				}
			}

			break;
		case 1://Y
			elem[0].height = src0->y;
			mag = src0->x * src0->x + src0->z * src0->z;
			if( mag > 0.0f ){
				elem[0].dist = (float)sqrt( mag );
			}else{
				elem[0].dist = 0.0f;
			}

			elem[1].height = src1->y;
			mag = src1->x * src1->x + src1->z * src1->z;
			if( mag > 0.0f ){
				elem[1].dist = (float)sqrt( mag );
			}else{
				elem[1].dist = 0.0f;
			}

			for( segno = 0; segno < m_lathe_seg; segno++ ){
				D3DXVECTOR3* dst0 = m_pointbuf2 + lineno * 2 * m_lathe_seg + segno;
				D3DXVECTOR3* dst1 = dst0 + m_lathe_seg;

				dst0->x = elem[0].dist * (float)cos( rad * segno );
				dst0->y = elem[0].height;
				dst0->z = elem[0].dist * (float)sin( rad * segno );

				dst1->x = elem[1].dist * (float)cos( rad * segno );
				dst1->y = elem[1].height;
				dst1->z = elem[1].dist * (float)sin( rad * segno );

				if( m_colorbuf2 ){
					D3DXVECTOR4* dstcol0 = m_colorbuf2 + lineno * 2 * m_lathe_seg + segno;
					D3DXVECTOR4* dstcol1 = dstcol0 + m_lathe_seg;

					*dstcol0 = *(m_colorbuf + v0);
					*dstcol1 = *(m_colorbuf + v1);
				}
			}
			break;
		case 2://Z
			elem[0].height = src0->z;
			mag = src0->x * src0->x + src0->y * src0->y;
			if( mag > 0.0f ){
				elem[0].dist = (float)sqrt( mag );
			}else{
				elem[0].dist = 0.0f;
			}

			elem[1].height = src1->z;
			mag = src1->x * src1->x + src1->y * src1->y;
			if( mag > 0.0f ){
				elem[1].dist = (float)sqrt( mag );
			}else{
				elem[1].dist = 0.0f;
			}

			for( segno = 0; segno < m_lathe_seg; segno++ ){
				D3DXVECTOR3* dst0 = m_pointbuf2 + lineno * 2 * m_lathe_seg + segno;
				D3DXVECTOR3* dst1 = dst0 + m_lathe_seg;

				dst0->x = elem[0].dist * (float)cos( rad * segno );
				dst0->y = elem[0].dist * (float)sin( rad * segno );
				dst0->z = elem[0].height;

				dst1->x = elem[1].dist * (float)cos( rad * segno );
				dst1->y = elem[1].dist * (float)sin( rad * segno );
				dst1->z = elem[1].height;

				if( m_colorbuf2 ){
					D3DXVECTOR4* dstcol0 = m_colorbuf2 + lineno * 2 * m_lathe_seg + segno;
					D3DXVECTOR4* dstcol1 = dstcol0 + m_lathe_seg;

					*dstcol0 = *(m_colorbuf + v0);
					*dstcol1 = *(m_colorbuf + v1);
				}
			}
			break;
		default:
			_ASSERT( 0 );
			break;
		}
	}


	for( lineno = 0; lineno < linenum; lineno++ ){ 
		for( segno = 0; segno < m_lathe_seg; segno++ ){
			CMQOFace* dstface = m_facebuf2 + m_lathe_seg * lineno + segno;
			dstface->m_pointnum = 4;

			if( segno != m_lathe_seg - 1 ){
				dstface->m_index[0] = lineno * 2 * m_lathe_seg + segno;
				dstface->m_index[1] = lineno * 2 * m_lathe_seg + m_lathe_seg + segno;
				dstface->m_index[2] = lineno * 2 * m_lathe_seg + m_lathe_seg + 1 + segno;
				dstface->m_index[3] = lineno * 2 * m_lathe_seg + 1 + segno;
			}else{
				dstface->m_index[0] = lineno * 2 * m_lathe_seg + segno;
				dstface->m_index[1] = lineno * 2 * m_lathe_seg + m_lathe_seg + segno;
				dstface->m_index[2] = lineno * 2 * m_lathe_seg + m_lathe_seg;
				dstface->m_index[3] = lineno * 2 * m_lathe_seg;
			}
			dstface->m_hasuv = 0;

		}
	}


	free( lineno2faceno );

	return 0;
}

int CMQOObject::MakeMirrorBuf()
{


	if( m_mirror == 2 ){
		CallF( FindConnectFace( 0 ), return 1 );//connectnum̃Zbg
		if( m_connectnum > 0 ){
			_ASSERT( m_connectface == 0 );
			m_connectface = new CMQOFace[ m_connectnum ];
			if( !m_connectface ){
				DbgOut( L"MQOObject : MakeMirrorBuf : connectface alloc error !!!\r\n" );
				_ASSERT( 0 );
				return 1;
			}

			CallF( FindConnectFace( 1 ), return 1 );//connectfacẽZbg
		}

	}else{
		m_connectnum = 0;
		m_connectface = 0;
	}
	

	int doconnect = 0;
	int axis;
	if( axis = (m_mirror_axis & 0x01) ){
		CallF( MakeMirrorPointAndFace( axis, doconnect ), return 1 );
		doconnect = 1;
	}

	if( axis = (m_mirror_axis & 0x02) ){
		CallF( MakeMirrorPointAndFace( axis, doconnect ), return 1 );
		doconnect = 1;
	}

	
	if( axis = (m_mirror_axis & 0x04) ){
		CallF( MakeMirrorPointAndFace( axis, doconnect ), return 1 );
		doconnect = 1;
	}



	return 0;
}

int CMQOObject::MakeMirrorPointAndFace( int axis, int doconnect )
{

	int befvertex, befface;

	if( m_vertex2 > 0 )
		befvertex = m_vertex2;
	else
		befvertex = m_vertex;
	
	if( m_face2 > 0 )
		befface = m_face2;
	else
		befface = m_face;

	m_vertex2 = befvertex * 2;
	
	if( doconnect == 0 )
		m_face2 = befface * 2 + m_connectnum;
	else
		m_face2 = befface * 2;

	if( m_pointbuf2 ){
		m_pointbuf2 = (D3DXVECTOR3*)realloc( m_pointbuf2, sizeof( D3DXVECTOR3 ) * m_vertex2 );
		if( !m_pointbuf2 ){
			DbgOut( L"MQOObject : MakeMirrorPointAndFace : pointbuf2 realloc error !!!\r\n" );
			_ASSERT( 0 );
			return 1;
		}
	}else{
		m_pointbuf2 = (D3DXVECTOR3*)malloc( sizeof( D3DXVECTOR3 ) * m_vertex2 );
		if( !m_pointbuf2 ){
			DbgOut( L"MQOObject : MakeMirrorPointAndFace : pointbuf2 alloc error !!!\r\n" );
			_ASSERT( 0 );
			return 1;
		}
		MoveMemory( m_pointbuf2, m_pointbuf, sizeof( D3DXVECTOR3 ) * befvertex );
	}

	if( m_colorbuf ){
		if( m_colorbuf2 ){
			m_colorbuf2 = (D3DXVECTOR4*)realloc( m_colorbuf2, sizeof( D3DXVECTOR4 ) * m_vertex2 );
			if( !m_colorbuf2 ){
				DbgOut( L"MQOObject : MakeMirrorPointAndFace : colorbuf2 realloc error !!!\r\n" );
				_ASSERT( 0 );
				return 1;
			}
		}else{
			m_colorbuf2 = (D3DXVECTOR4*)malloc( sizeof( D3DXVECTOR4 ) * m_vertex2 );
			if( !m_colorbuf2 ){
				DbgOut( L"MQOObject : MakeMirrorPointAndFace : colorbuf2 alloc error !!!\r\n" );
				_ASSERT( 0 );
				return 1;
			}
			MoveMemory( m_colorbuf2, m_colorbuf, sizeof( D3DXVECTOR4 ) * befvertex );
		}
	}

	if( m_facebuf2 ){
		CMQOFace* newface;
		newface = new CMQOFace[ m_face2 ];
		if( !newface ){
			DbgOut( L"MQOObject : MakeMirrorPointAndFace : newface alloc error !!!\r\n" );
			_ASSERT( 0 );
			return 1;
		}
		MoveMemory( newface, m_facebuf2, sizeof( CMQOFace ) * befface );
		delete [] m_facebuf2;
		m_facebuf2 = newface;

	}else{
		m_facebuf2 = new CMQOFace[ m_face2 ];
		if( !m_facebuf2 ){
			DbgOut( L"MQOObject : MakeMirrorPointAndFace : facebuf2 alloc error !!!\r\n" );
			_ASSERT( 0 );
			return 1;
		}
		MoveMemory( m_facebuf2, m_facebuf, sizeof( CMQOFace ) * befface );
	}


	int vertno;
	D3DXVECTOR3* srcvec;
	D3DXVECTOR3* dstvec;
	switch( axis ){
	case 1://X
		for( vertno = 0; vertno < befvertex; vertno++ ){
			srcvec = m_pointbuf2 + vertno;
			dstvec = m_pointbuf2 + vertno + befvertex;
			
			dstvec->x = -srcvec->x;
			dstvec->y = srcvec->y;
			dstvec->z = srcvec->z;
		}
		break;
	case 2://Y
		for( vertno = 0; vertno < befvertex; vertno++ ){
			srcvec = m_pointbuf2 + vertno;
			dstvec = m_pointbuf2 + vertno + befvertex;
			
			dstvec->x = srcvec->x;
			dstvec->y = -srcvec->y;
			dstvec->z = srcvec->z;
		}
		break;
	case 4://Z
		for( vertno = 0; vertno < befvertex; vertno++ ){
			srcvec = m_pointbuf2 + vertno;
			dstvec = m_pointbuf2 + vertno + befvertex;
			
			dstvec->x = srcvec->x;
			dstvec->y = srcvec->y;
			dstvec->z = -srcvec->z;
		}

		break;
	default:
		_ASSERT( 0 );
		break;
	}

	if( m_colorbuf2 ){
		D3DXVECTOR4* srccol;
		D3DXVECTOR4* dstcol;
		for( vertno = 0; vertno < befvertex; vertno++ ){
			srccol = m_colorbuf2 + vertno;
			dstcol = m_colorbuf2 + vertno + befvertex;
			
			*dstcol = *srccol;
		}
	}


	int faceno;
	CMQOFace* srcface;
	CMQOFace* dstface;
	for( faceno = 0; faceno < befface; faceno++ ){
		srcface = m_facebuf2 + faceno;
		dstface = m_facebuf2 + faceno + befface;

		CallF( dstface->SetInvFace( srcface, befvertex ), return 1 );
	}


	if( (doconnect == 0) && m_connectface ){
		int srcno = 0;
		for( faceno = befface * 2; faceno < m_face2; faceno++ ){
			srcface = m_connectface + srcno;
			dstface = m_facebuf2 + faceno;
			
			dstface->m_pointnum = 4;
			dstface->m_materialno = srcface->m_materialno;
			
			dstface->m_index[0] = srcface->m_index[0];//Όɕ`悷悤ɃZbgB
			dstface->m_index[1] = srcface->m_index[0] + befvertex;
			dstface->m_index[2] = srcface->m_index[1] + befvertex;
			dstface->m_index[3] = srcface->m_index[1];

			dstface->m_hasuv = srcface->m_hasuv;

			dstface->m_uv[0] = srcface->m_uv[0];
			dstface->m_uv[1] = srcface->m_uv[0];
			dstface->m_uv[2] = srcface->m_uv[1];
			dstface->m_uv[3] = srcface->m_uv[1];

			dstface->m_col[0] = srcface->m_col[0];
			dstface->m_col[1] = srcface->m_col[0];
			dstface->m_col[2] = srcface->m_col[1];
			dstface->m_col[3] = srcface->m_col[1];

			srcno++;
		}
	}


	return 0;
}


int CMQOObject::FindConnectFace( int issetface )
{

	int conno = 0;
	int faceno, chkno;
	int pointnum;
	int findflag[4];
	int findall;
	CMQOFace* curface;
	CMQOFace* chkface;
	for( faceno = 0; faceno < m_face; faceno++ ){
		curface = m_facebuf + faceno;
		pointnum = curface->m_pointnum;
		if( pointnum <= 2 )
			continue;

		ZeroMemory( findflag, sizeof( int ) * 4 );

		findall = 0;
		for( chkno = 0; chkno < m_face; chkno++ ){
			if( chkno == faceno )
				continue;
			chkface = m_facebuf + chkno;

			if( chkface->m_pointnum <= 2 )
				continue;

			CallF( curface->CheckSameLine( chkface, findflag ), return 1 );

			int flagsum = 0;
			int sumno;
			for( sumno = 0; sumno < pointnum; sumno++ ){
				flagsum += findflag[sumno];
			}
			if( flagsum == pointnum ){
				findall = 1;
				break;
			}
		}

		if( findall == 0 ){
			int i;
			for( i = 0; i < pointnum; i++ ){
				if( findflag[i] == 0 ){//LĂȂ
					int distok = 1;
					distok = CheckMirrorDis( m_pointbuf, curface, i, pointnum );
					if( distok ){
						if( issetface ){
							(m_connectface + conno)->m_pointnum = 2;
							(m_connectface + conno)->m_materialno = curface->m_materialno;
							(m_connectface + conno)->m_hasuv = curface->m_hasuv;

							(m_connectface + conno)->m_index[0] = curface->m_index[i];
							(m_connectface + conno)->m_uv[0] = curface->m_uv[i];
							if( i != (pointnum - 1) ){
								(m_connectface + conno)->m_index[1] = curface->m_index[i+1];
								(m_connectface + conno)->m_uv[1] = curface->m_uv[i+1];
							}else{
								(m_connectface + conno)->m_index[1] = curface->m_index[0];
								(m_connectface + conno)->m_uv[1] = curface->m_uv[0];
							}

						}
						conno++;
					}
				}
			}

		}

	}

	if( issetface == 0 ){
		m_connectnum = conno;
	}


	return 0;
}


int CMQOObject::CheckMirrorDis( D3DXVECTOR3* pbuf, CMQOFace* fbuf, int lno, int pnum )
{
	if( m_issetmirror_dis == 0 ){
		return 1;
	}else{
		int axisx, axisy, axisz;
		axisx = m_mirror_axis & 0x01;
		axisy = m_mirror_axis & 0x02;
		axisz = m_mirror_axis & 0x04;

		int i0, i1;
		i0 = lno;
		if( lno == pnum - 1 ){
			i1 = 0;
		}else{
			i1 = lno + 1;
		}
		D3DXVECTOR3* v0;
		D3DXVECTOR3* v1;
		v0 = pbuf + fbuf->m_index[i0];
		v1 = pbuf + fbuf->m_index[i1];

		float distx0, distx1;
		int chkx = 1;
		if( axisx ){
			distx0 = v0->x * 2.0f;
			distx1 = v1->x * 2.0f;

			if( (distx0 > m_mirror_dis) || (distx1 > m_mirror_dis) ){
				chkx = 0;
			}
		}


		float disty0, disty1;
		int chky = 1;
		if( axisy ){
			disty0 = v0->y * 2.0f;
			disty1 = v1->y * 2.0f;

			if( (disty0 > m_mirror_dis) || (disty1 > m_mirror_dis) ){
				chky = 0;
			}
		}


		float distz0, distz1;
		int chkz = 1;
		if( axisz ){
			distz0 = v0->z * 2.0f;
			distz1 = v1->z * 2.0f;

			if( (distz0 > m_mirror_dis) || (distz1 > m_mirror_dis) ){
				chkz = 0;
			}
		}


		if( (chkx == 1) && (chky == 1) && (chkz == 1) ){
			return 1;
		}else{
			return 0;
		}

	}

}

int CMQOObject::MultMat( D3DXMATRIX multmat )
{
	m_multmat = multmat;
	return 0;

}



int CMQOObject::Multiple( float multiple )
{
	D3DXMatrixIdentity( &m_multmat );
	m_multmat._11 = multiple;
	m_multmat._22 = multiple;
	m_multmat._33 = -multiple;

	return 0;
}
int CMQOObject::InitFaceDirtyFlag()
{
	int fno;
	CMQOFace* curface;
	for( fno = 0; fno < m_face; fno++ ){
		curface = m_facebuf + fno;

		curface->m_dirtyflag = 0;
	}

	return 0;
}

